# -*- coding: utf-8 -*-
# Author: Runsheng Xu <rxx3386@ucla.edu>
# License: TDG-Attribution-NonCommercial-NoDistrib


import argparse
import os
import statistics
import numpy as np

import sys
root_path = os.path.abspath(__file__)
root_path = '/'.join(root_path.split('/')[:-3])
sys.path.append(root_path)

import torch
from torch.utils.data import DataLoader, Subset
from tensorboardX import SummaryWriter

import opencood.hypes_yaml.yaml_utils as yaml_utils
from opencood.tools import train_utils
from opencood.data_utils.datasets import build_dataset

#from icecream import ic


def train_parser():
    parser = argparse.ArgumentParser(description="synthetic data generation")
    parser.add_argument("--hypes_yaml", "-y", type=str, required=True,
                        help='data generation yaml file needed ')
    parser.add_argument('--model_dir', default='',
                        help='Continued training path')
    parser.add_argument('--fusion_method', '-f', default="intermediate",
                        help='passed to inference.')
    parser.add_argument('--epoch', default=None,
                        help='continue traning with epoch.')
    parser.add_argument('--start_from_best', default=False,
                        help='continue traning with the best epoch.')
    opt = parser.parse_args()
    return opt


def main():
    opt = train_parser()
    hypes = yaml_utils.load_yaml(opt.hypes_yaml, opt)

    print('Dataset Building')
    opencood_train_dataset = build_dataset(hypes, visualize=False, train=True)
    opencood_validate_dataset = build_dataset(hypes,
                                              visualize=False,
                                              train=False)

    train_loader = DataLoader(opencood_train_dataset,
                              batch_size=hypes['train_params']['batch_size'],
                              num_workers=0,
                              collate_fn=opencood_train_dataset.collate_batch_train,
                              shuffle=True,
                              pin_memory=True,
                              drop_last=True,
                              prefetch_factor=2)
    val_loader = DataLoader(opencood_validate_dataset,
                            batch_size=hypes['train_params']['batch_size'],
                            num_workers=0,
                            collate_fn=opencood_train_dataset.collate_batch_train,
                            shuffle=True,
                            pin_memory=True,
                            drop_last=True,
                            prefetch_factor=2)

    print('Creating Model')
    model = train_utils.create_model(hypes)
    #print(model)
    total = sum([param.nelement() for param in model.parameters()])
    print("Number of parameter: %d" % (total))

    device = torch.device(hypes['device'])
    # device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
    
    # record lowest validation loss checkpoint.
    lowest_val_loss = 1e5
    lowest_val_epoch = -1


    # define the loss
    criterion = train_utils.create_loss(hypes)

    # optimizer setup
    optimizer = train_utils.setup_optimizer(hypes, model)
    # lr scheduler setup


    # teacher_model
    global_step = 0
    alpha = 0.9996   # ema_decay

    teacher_model = train_utils.create_model(hypes)
    # load init model
    teacher_model_path = hypes['teacher_model'] if 'teacher_model' in hypes else None
    if teacher_model_path is None or not os.path.exists(teacher_model_path):
        raise IOError('Teacher model file is not exist.')
    model.load_state_dict(torch.load(teacher_model_path, map_location='cpu'), strict=False)
    teacher_model.load_state_dict(torch.load(teacher_model_path, map_location='cpu'), strict=False)
    for param in teacher_model.parameters():
        param.detach_()

    # if we want to train from last checkpoint.
    if opt.model_dir:
        saved_path = opt.model_dir
        init_epoch, model = train_utils.load_model(saved_path, model, opt.epoch, start_from_best=opt.start_from_best)
        lowest_val_epoch = init_epoch
        scheduler = train_utils.setup_lr_schedular(hypes, optimizer, init_epoch=init_epoch)

        global_step = (init_epoch - 1)*(len(train_loader)//hypes['train_params']['batch_size']+np.ceil(len(train_loader)%hypes['train_params']['batch_size']))
    else:
        init_epoch = 0
        # if we train the model from scratch, we need to create a folder
        # to save the model,
        saved_path = train_utils.setup_train(hypes)
        scheduler = train_utils.setup_lr_schedular(hypes, optimizer)

    # we assume gpu is necessary
    if torch.cuda.is_available():
        model.to(device)
        teacher_model.to(device)
        #global_step = torch.Tensor([global_step]).to(device)
        #alpha = torch.Tensor([alpha]).to(device)
    # record training
    writer = SummaryWriter(saved_path)

    print('Training start')
    epoches = hypes['train_params']['epoches']
    supervise_single_flag = True if 'supervise_single' in hypes['train_params'] and hypes['train_params']['supervise_single'] else False
    # used to help schedule learning rate


    for epoch in range(init_epoch, max(epoches, init_epoch)):
        lrs = []
        for param_group in optimizer.param_groups:
            print('learning rate %f' % param_group["lr"])
            lrs.append(param_group["lr"])
        
        for i, batch_data in enumerate(train_loader):
            if batch_data is None:
                continue
            # the model will be evaluation mode during validation
            model.train()
            model.zero_grad()
            optimizer.zero_grad()

            batch_data = train_utils.to_device(batch_data, device)
            batch_data['ego']['epoch'] = epoch
            batch_data['ego']['mode'] = 'train'

            ouput_dict = model(batch_data['ego'])
            
            # final_loss = criterion(ouput_dict, batch_data['ego']['label_dict'])
            # criterion.logging(epoch, i, len(train_loader), writer)
            final_loss = criterion._forward(ouput_dict, batch_data['ego']['label_dict'])
            criterion._logging(epoch, i, len(train_loader), writer)

            if supervise_single_flag:
                #final_loss += criterion(ouput_dict, batch_data['ego']['label_dict_single'], suffix="_single")
                #criterion.logging(epoch, i, len(train_loader), writer, suffix="_single")
                final_loss += criterion._forward(ouput_dict, batch_data['ego']['label_dict_single'], suffix="_single")
                criterion._logging(epoch, i, len(train_loader), writer, suffix="_single")

            with open(os.path.join(saved_path, 'train_loss.txt'), 'a+') as f:
                msg = 'Epoch[{}], iter[{}/{}], loss[{}]. \n'.format(epoch, i, len(train_loader), final_loss)
                f.write(msg)
            
            # back-propagation
            final_loss.backward()
            optimizer.step()
            torch.cuda.empty_cache()

            # teacher model update
            global_step += 1
            teacher_model.train()
            # Use the true average until the exponential average is more correct
            alpha = min(1 - 1 / (global_step + 1), alpha)
            for ema_param, param in zip(teacher_model.parameters(), model.parameters()):
                # ema_param.data.mul_(alpha).add_(1 - alpha, param.data)
                ema_param.data.mul_(alpha).add_(param.data, alpha=1 - alpha)
            

        if epoch % hypes['train_params']['eval_freq'] == 0:
            valid_ave_loss = []

            with torch.no_grad():
                for i, batch_data in enumerate(val_loader):
                    if batch_data is None:
                        continue
                    teacher_model.zero_grad()
                    optimizer.zero_grad()
                    teacher_model.eval()

                    batch_data = train_utils.to_device(batch_data, device)
                    batch_data['ego']['epoch'] = epoch
                    batch_data['ego']['mode'] = 'validate'
                    ouput_dict = teacher_model(batch_data['ego'])

                    final_loss = criterion(ouput_dict,
                                           batch_data['ego']['label_dict'])
                    valid_ave_loss.append(final_loss.item())

            valid_ave_loss = statistics.mean(valid_ave_loss)
            print('At epoch %d, the validation loss is %f' % (epoch,
                                                              valid_ave_loss))
            writer.add_scalar('Validate_Loss', valid_ave_loss, epoch)

            with open(os.path.join(saved_path, 'validation_loss.txt'), 'a+') as f:
                msg = 'Epoch[{}], loss[{}], lr[{}]. \n'.format(epoch, valid_ave_loss, lrs)
                f.write(msg)

            # lowest val loss
            if valid_ave_loss < lowest_val_loss:
                lowest_val_loss = valid_ave_loss
                torch.save(teacher_model.state_dict(),
                       os.path.join(saved_path,
                                    'net_epoch_bestval_at%d.pth' % (epoch + 1)))
                if lowest_val_epoch != -1 and os.path.exists(os.path.join(saved_path,
                                    'net_epoch_bestval_at%d.pth' % (lowest_val_epoch))):
                    os.remove(os.path.join(saved_path,
                                    'net_epoch_bestval_at%d.pth' % (lowest_val_epoch)))
                lowest_val_epoch = epoch + 1

        if epoch % hypes['train_params']['save_freq'] == 0:
            torch.save(teacher_model.state_dict(),
                       os.path.join(saved_path,
                                    'net_epoch%d.pth' % (epoch + 1)))
        scheduler.step(epoch)

    print('Training Finished, checkpoints saved to %s' % saved_path)

    run_test = True
    if run_test:
        fusion_method = opt.fusion_method
        cmd = f"python opencood/tools/inference_where2comm.py --model_dir {saved_path} --fusion_method {fusion_method}"
        print(f"Running command: {cmd}")
        os.system(cmd)

if __name__ == '__main__':
    main()
